#include <string>
using namespace std;

#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <dirent.h>

#include "debug.hpp"
#include "utility.hpp"
#include "config.hpp"
#include "outstreams.hpp"
#include "SC2Map.hpp"
#include "SC2MapAggregator.hpp"



// controls set by options: defaults
static bool pauseTerminal   = true;
static bool dirRecurse      = false;
static bool renderTerrain   = false;
static bool renderPathing   = false;
static bool renderBases     = false;
static bool renderOpenness  = false;
static bool renderShortest  = false;
static bool renderInfluence = false;
static bool renderSummary   = false;
static bool writeCSVpermap  = false;
static bool writeCSVaggr    = false;
//static bool defaultOutput   = true;


void optsFail();
void statFail( string* fileFullPath );

void tryFileOrDir( string* fileFullPath, bool first, bool recurse );
void tryFile     ( string* fileFullPath );


void usage()
{
  printMessage( "\n" );
  printMessage( "usage: sc2mapanalyzer FILES\n" );
  printMessage( "  If FILE is a directory, all maps inside are analyzed.\n" );
  printMessage( "\n" );
  printMessage( "Use -r to recursively explore all directory arguments.\n" );
  printMessage( "Use -p to prevent the command window from staying open after analysis.\n" );
  printMessage( "Use -h to print a brief description of available options.\n" );
  printMessage( "\n" );
  printMessage( "For the complete manual, visit:\n" );
  printMessage( "http://www.sc2mapster.com/assets/sc2-map-analyzer\n" );
}


void help()
{
  printMessage( "\n" );
  printMessage( "usage: sc2mapanalyzer FILES\n" );
  printMessage( "  If FILE is a directory, all maps inside are analyzed.\n" );
  printMessage( "\n" );
  printMessage( "-h  Display this help message.\n" );
  printMessage( "\n" );
  printMessage( "-p  Prevent the terminal from pausing open after analysis.\n" );
  printMessage( "\n" );
  printMessage( "-r  Recursively search for maps in any directory arguments.\n" );
  printMessage( "\n" );
  printMessage( "-m  Set the openness render max, default is 15.0 (in map cells).\n" );
  printMessage( "\n" );
  printMessage( "-g -c -a  Set ground/cliff-walk/air influence weights, should add up to 1.\n" );
  printMessage( "\n" );
  printMessage( "-d  Set the openness neighborhood radius, default is 12.0 (in map cells).\n" );
  printMessage( "\n" );
  printMessage( "-t  Render PNG of terrain only.\n" );
  printMessage( "\n" );
  printMessage( "-n  Render PNGs of pathing data.\n" );
  printMessage( "\n" );
  printMessage( "-b  Render PNG of bases.\n" );
  printMessage( "\n" );
  printMessage( "-o  Render PNG of openness.\n" );
  printMessage( "\n" );
  printMessage( "-s  Render PNGs of shortest paths.\n" );
  printMessage( "\n" );
  printMessage( "-i  Render PNGs of influence calculations.\n" );
  printMessage( "\n" );
  printMessage( "-z  Render a summary PNG.\n" );
  printMessage( "\n" );
  printMessage( "-v  Write a CSV summarizing all maps.\n" );
  printMessage( "\n" );
  printMessage( "-w  Write a CSV for each map's details.\n" );
  printMessage( "\n" );
  printMessage( "-e  Enable all outputs described above.\n" );
  printMessage( "\n" );
}


void AtExit()
{
  // flush output
  fflush( stdout );
  fflush( stderr );

  printOutstreamStatus();

  // make sure to hold terminal open so
  // user can read output
  if( pauseTerminal )
  {
    //printMessage( "\nTo prevent this pause at termination, pass option -p\n" );
    system( "pause" );
  }
}


static string           toolPath;
//static string           outputPath;
static SC2MapAggregator mapAggregator;


int main( int argc, char** argv )
{
  atexit( &AtExit );


  if( SC2Map::pngwriterTest() != 0 )
  {
    printError( "Cannot load resources for PNGwriter: you may need to run this program as Administrator.\n" );
    exit( -1 );
  }

  string toolFullName( argv[0] );

  string::size_type locBSlash = toolFullName.rfind( "\\" );
  toolPath.assign( toolFullName, 0, locBSlash + 1 );
  formatPath( &toolPath );

  printMessage( "SC2 Map Analyzer v%s, algorithms v%s\n  tool path: %s\n\n",
                QUOTEMACRO( VEXE ),
                QUOTEMACRO( VALG ),
                toolPath.data()
              );

  initConfigReading();
  readConfigFiles( false, NULL, &configInternal );

  readConfigFiles( true, &toolPath, &configUserGlobal );


  /*
  // from unistd.h for getopt
  opterr = 0;

  int c;
  while( (c = getopt( argc, argv, "hprm:g:c:a:d:tnbosivwezj:" )) != -1 )
  {
    switch( c )
    {
      case 'j':
        // secret dev options
        if( strcmp( optarg, "debugMapInfo" ) == 0 ) {
          debugMapInfo = 1;

        } else if( strcmp( optarg, "debugObjects" ) == 0 ) {
          debugObjects = 1;
        }
        break;

      case 'h':
        help();
        exit( 0 );
        break;

      case 'p':
        pauseTerminal = false;
        break;

      case 'r':
        dirRecurse = true;
        break;

      case 'm':
        SC2Map::setOpennessRenderMax( atof( optarg ) );
        break;

      case 'g':
        SC2Map::setInfluenceWeightGround( atof( optarg ) );
        break;

      case 'c':
        SC2Map::setInfluenceWeightCWalk( atof( optarg ) );
        break;

      case 'a':
        SC2Map::setInfluenceWeightAir( atof( optarg ) );
        break;

      case 'd':
        SC2Map::setOpennessNeighborhoodRadius( atof( optarg ) );
        break;

      case 't':
        renderTerrain = true;
        defaultOutput = false;
        break;

      case 'n':
        renderPathing = true;
        defaultOutput = false;
        break;

      case 'b':
        renderBases   = true;
        defaultOutput = false;
        break;

      case 'o':
        renderOpenness = true;
        defaultOutput  = false;
        break;

      case 's':
        renderShortest = true;
        defaultOutput  = false;
        break;

      case 'i':
        renderInfluence = true;
        defaultOutput   = false;
        break;

      case 'z':
        renderSummary = true;
        defaultOutput = false;
        break;

      case 'v':
        writeCSVaggr  = true;
        defaultOutput = false;
        break;

      case 'w':
        writeCSVpermap  = true;
        defaultOutput = false;
        break;

      case 'e':
        renderTerrain   = true;
        renderPathing   = true;
        renderBases     = true;
        renderOpenness  = true;
        renderShortest  = true;
        renderInfluence = true;
        renderSummary   = true;
        writeCSVaggr    = true;
        writeCSVpermap    = true;
        defaultOutput   = false;
        break;

      case ':':
      case '?':
        // bad options, inform user
        optsFail();
        // no break, fall through

      default:
        usage();
        exit( -1 );
    }
  }


  // if no output was specified, enable
  // the default output
  if( defaultOutput )
  {
    renderTerrain   = false;
    renderPathing   = false;
    renderBases     = true;
    renderOpenness  = true;
    renderShortest  = true;
    renderInfluence = true;
    renderSummary   = true;
    writeCSVaggr    = true;
    writeCSVpermap    = true;
  }
  */

  writeCSVaggr = getGlobalOutputOption( "writeCSVaggr" );

  /*
  // if, after options, there are no arguments
  // then search the current directory for maps
  if( optind == argc )
  {
    printMessage( "No file arguments, searching current directory for maps...\n\n" );
    string currentDir( "." );
    tryFileOrDir( &currentDir, true );
  }

  // process non-option arguments
  for( int i = optind; i < argc; ++i )
  {
    string fileFullPath( argv[i] );
    tryFileOrDir( &fileFullPath, true );
  }
  */

  if( configUserGlobal.toAnalyze.empty() &&
      configUserGlobal.toAnalyzeRecurse.empty() )
  {
    printMessage( "No file arguments, searching current directory for maps...\n\n" );
    string currentDir( "." );
    tryFileOrDir( &currentDir, true, true );
  }

  for( list<string>::iterator argItr = configUserGlobal.toAnalyze.begin();
       argItr != configUserGlobal.toAnalyze.end();
       ++argItr )
  {
    tryFileOrDir( &(*argItr), true, false );
  }

  for( list<string>::iterator argItr = configUserGlobal.toAnalyzeRecurse.begin();
       argItr != configUserGlobal.toAnalyzeRecurse.end();
       ++argItr )
  {
    tryFileOrDir( &(*argItr), true, true );
  }



  if( writeCSVaggr )
  {
    printMessage( "Calculating aggregate statistics...\n" );
    mapAggregator.buildOutputColumns();

    if( configUserGlobal.outputPath.empty() )
    {
      mapAggregator.writeToCSV( &toolPath );
    } else {
      mapAggregator.writeToCSV( &(configUserGlobal.outputPath) );
    }
  }

  return 0;
}


void optsFail()
{
  if( optopt == 'm' ||
      optopt == 'g' ||
      optopt == 'c' ||
      optopt == 'a' ||
      optopt == 'd' )
  {
    printError( "Option -%c requires an argument.\n", optopt );
  }

  else if (isprint (optopt))
  {
    printError( "Unknown option `-%c'.\n", optopt );

  } else {
    fprintf (stderr,
           "Unknown option character `\\x%x'.\n",
           optopt);
         }
}



void statFail( string* fileFullPath )
{
  const char* filename = fileFullPath->data();

  switch( errno )
  {
    case EACCES:
      printWarning( "Permission denied for %s, skipping...\n",
                    filename );
      break;

    case EBADF:
    case EFAULT:
      printWarning( "%s is an invalid file, skipping...\n",
                    filename );
      break;

    case ENAMETOOLONG:
      printWarning( "Path and filename %s is too long, skipping...\n",
                    filename );
      break;

    case ENOENT:
    case ENOTDIR:
      printWarning( "A component of the path %s does not exist or is not a directory, skipping...\n",
                    filename );
      break;

    case ENOMEM:
      printError( "Not enough system memory to continue.\n" );
      exit( -1 );
      break;

    default:
      printWarning( "Unknown problem trying to access %s, skipping...\n",
                    filename );
  }
}


// find out if the non-option argument is a file or
// a directory, and for directories decide if we should
// recursively explore them if the flag is set
void tryFileOrDir( string* fileFullPathIn, bool first, bool recurse )
{
  // make a copy of the path and filename for mutation
  string fileFullPath( *fileFullPathIn );

  formatPath( &fileFullPath );

  // find out what kind of file we have
  struct stat fileStatus;
  if( stat( fileFullPath.data(), &fileStatus ) < 0 )
  {
    statFail( &fileFullPath );
    return;
  }

  // only deal with regular files or directories,
  // ignore other file types (or "modes")
  if( S_ISREG( fileStatus.st_mode ) )
  {
    tryFile( &fileFullPath );


  } else if( S_ISDIR( fileStatus.st_mode ) ) {

    if( !first && !recurse )
    {
      return;
    }

    DIR* dir = opendir( fileFullPath.data() );
    if( dir != NULL )
    {
      struct dirent* de;
      while( de = readdir( dir ) )
      {
        // ignore the current and up-one directories, . and ..
        if( strcmp( de->d_name, "."  ) == 0 ||
            strcmp( de->d_name, ".." ) == 0 )
        {
          continue;
        }

        // make a path out of this directory name
        string deeperFile( fileFullPath );
        deeperFile.append( "\\" );
        deeperFile.append( de->d_name );

        // only continue to recurse beyond the first level
        // if the user passed recursive option
        tryFileOrDir( &deeperFile, false, recurse );
      }
      closedir( dir );

    } else {
      printWarning( "Could not open directory %s.\n", fileFullPath.data() );
    }

  } else {
    printWarning( "%s is not a regular file or directory, skipping...", fileFullPath.data() );
  }
}



// found a regular file, see if it has a map
// extension and then try to open the archive
void tryFile( string* fileFullPath )
{
  string ext1( ".s2ma" );
  string ext2( ".SC2Map" );

  string::size_type loc1 = fileFullPath->rfind( ext1 );
  string::size_type loc2 = fileFullPath->rfind( ext2 );

  // if either search for an extension doesn't come
  // up with the location where the extension should be...
  if( loc1 != fileFullPath->size() - ext1.size() &&
      loc2 != fileFullPath->size() - ext2.size() )
  {
    // not a Starcraft 2 Map Archive...
    return;
  }

  // find last dot, last back slash
  string::size_type locDot    = fileFullPath->rfind( "." );
  string::size_type locBSlash = fileFullPath->rfind( "\\" );

  string path( *fileFullPath, 0,             locBSlash + 1                 );
  string file( *fileFullPath, locBSlash + 1, locDot - locBSlash - 1        );
  string ext ( *fileFullPath, locDot,        fileFullPath->size() - locDot );

  string localOutputPath( path );

  // consider switching to the globally specified
  // output path, if it exists AND is valid
  if( !configUserGlobal.outputPath.empty() )
  {
    struct stat dirStatus;
    if( stat( configUserGlobal.outputPath.data(), &dirStatus ) >= 0 )
    {
      if( S_ISDIR( dirStatus.st_mode ) )
      {
        localOutputPath.assign( configUserGlobal.outputPath );
      }
    }
  }


  string fileWithExt( file );
  fileWithExt.append( ext );

  SC2Map* sc2map = new SC2Map( toolPath,
                               localOutputPath,
                               path,
                               file,
                               fileWithExt );

  printMessage( "\nOpening %s...\n", fileFullPath->data() );

  readConfigFiles( false, &path, &(sc2map->configUserLocal) );

  // consider switching to the locally specified
  // output path, if it exists AND is valid
  if( !sc2map->configUserLocal.outputPath.empty() )
  {
    struct stat dirStatus;
    if( stat( sc2map->configUserLocal.outputPath.data(), &dirStatus ) >= 0 )
    {
      if( S_ISDIR( dirStatus.st_mode ) )
      {
        sc2map->outputPath = sc2map->configUserLocal.outputPath;
      }
    }
  }

  renderTerrain   = sc2map->getOutputOption( "renderTerrain"   );
  renderPathing   = sc2map->getOutputOption( "renderPathing"   );
  renderBases     = sc2map->getOutputOption( "renderBases"     );
  renderOpenness  = sc2map->getOutputOption( "renderOpenness"  );
  renderShortest  = sc2map->getOutputOption( "renderShortest"  );
  renderInfluence = sc2map->getOutputOption( "renderInfluence" );
  renderSummary   = sc2map->getOutputOption( "renderSummary"   );
  writeCSVpermap  = sc2map->getOutputOption( "writeCSVpermap"  );

  if( sc2map->readMap() < 0 )
  {
    printWarning( "Could not read required map files for %s, skipping\n\n", file.data() );
    return;
  }

  printMessage( "Prepping analysis,\n" );

  sc2map->countPathableCells();

  printMessage( "." );

  sc2map->prepShortestPaths();

  printMessage( "." );

  sc2map->identifyBases();

  printMessage( "." );

  sc2map->computeOpenness();

  printMessage( "." );

  sc2map->analyzeBases();

  printMessage( "\nAnalyzing and generating output,\n" );
  printMessage( "  Map-specific output in [%s]\n", sc2map->outputPath.data() );

  if( renderTerrain )
  {
    sc2map->renderImageTerrainOnly();
  }

  printMessage( "." );

  if( renderPathing )
  {
    sc2map->renderImagesPathing();
  }

  printMessage( "." );

  if( renderBases )
  {
    sc2map->renderImageBases();
  }

  printMessage( "." );

  if( renderOpenness )
  {
    sc2map->renderImageOpenness();
  }

  printMessage( "." );

  if( renderShortest )
  {
    sc2map->renderImagesShortestPaths();
  }

  printMessage( "." );

  if( renderInfluence )
  {
    sc2map->renderImagesInfluence();
  }

  printMessage( "." );

  if( renderSummary )
  {
    sc2map->renderImageSummary();
  }

  printMessage( "." );

  if( writeCSVpermap )
  {
    sc2map->writeToCSV();
  }

  printMessage( "." );

  if( writeCSVaggr )
  {
    mapAggregator.aggregate( sc2map );
  }

  printMessage( "\n\n" );


  // normally the guts of this function
  // are commented out, uncomment for testing
  sc2map->renderTestImages();


  delete sc2map;
}
